/* validate L(#)-convex function
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "laminar.h"

/* returns the minimum value of func(x)    */
/* INPUT  func    : pointer to the function */
/* INPUT  p[],q[] : 2point */
/* work   p2[],q2[] :   */
static int validate(int dim, double func(int,int*),
                    int p[], int q[], int p2[], int q2[]) {
    double fp1, fq1, fp2, fq2;
    int i;

    /* Don't validate if outside of domain */
    oracle_call++;
    fp1 = func(dim, p); if (fp1 == PLUS_INFINITY) return TRUE;
    oracle_call++;
    fq1 = func(dim, q); if (fq1 == PLUS_INFINITY) return TRUE;
    /* make two middle points */
    for (i=0; i<dim; i++) {
        int sum = p[i] + q[i];
        if (sum % 2 == 0) {
            p2[i] = q2[i] = sum / 2;
        } else {
            p2[i] = (sum-1)/2;
            q2[i] = (sum+1)/2;
        }
    }
   /* return FALSE if either of two middle points is outside of domain */
    oracle_call++;
    fp2 = func(dim, p2); if (fp2 == PLUS_INFINITY) return FALSE;
    oracle_call++;
    fq2 = func(dim, q2); if (fq2 == PLUS_INFINITY) return FALSE;
    return (fp1+fq1 >= fp2+fq2);
}

static int next(int dim, int x[], int lower[], int upper[], int step) {
    int i = 0;

    while ((x[i]+=step) > upper[i]) {
        x[i] = lower[i];
        if (++i >= dim) return FALSE; /* end mark */
    }
    return TRUE;
}

/* decide if the function is L-convex or not */
int lconv_validate(int dim, double func(int,int*),
		   int lower[], int upper[]) {
    int ret = TRUE, step;
    int *p = ivector_new(dim);
    int *q = ivector_new(dim);
    int *p2 = ivector_new(dim);
    int *q2 = ivector_new(dim);

    ivector_copy(dim, p, lower);
    ivector_copy(dim, q, lower);

    step = upper[0]-lower[0];
    while ((step & (step-1)) != 0) step &= step - 1;

    for (step--; step>=1; step/=2) {
        do {
            do {
#ifdef DEBUG_DISP
                printf("p=");ivector_print(dim, p);
                printf(", q=");ivector_print(dim, q);
                printf("\n");
#endif /* DEBUG_DISP */
                if (!validate(dim, func, p, q, p2, q2)) {
                    ret = FALSE;
                    printf("p=");ivector_print(dim, p);
                    printf(", q=");ivector_print(dim, q);
                    printf(", step=%d\n", step);
                }
            } while (ret && next(dim, q, lower, upper, step));
        } while (ret && next(dim, p, lower, upper, step));
    }
    free(p);
    free(q);
    free(p2);
    free(q2);
    return ret;
}


/* f1(x) = 2*x0^2 + 3*x1^2 - 2*x0*x1 */
/* 2次形式で書くと Hessian は [2, -1; -1, 3] */
static double f1(int dim, int x[]) {
    double r = 0;

    assert(dim == 2);  /* check dimension */
    r +=  2 * x[0] * x[0];
    r +=  3 * x[1] * x[1];
    r += -2 * x[0] * x[1];
    return r;
}

/* 一部を除いて f1() と同じ */
static double f2(int dim, int x[]) {
    if (0<=x[0] && x[0]<5 && 0<=x[1] && x[1]<5) return 100;
    return f1(dim, x);
}

/* 層型の M#-convex */
static double f3(int dim, int x[]) {
    static int v[] = {
        2,40,90, 1,2,   -1,/* y=(x1+x2),    f += 2y^2+40y+233 */
        50,-2,0, 1,     -1,/* y=(x1),       f += 50y^2-2y */
        3,20,0,  1,2,3, -1,/* y=(x1+x2+x3), f += 3y^2+20y */
        1,1,0,   0,     -1,/* y=(x0),       f += y^2 + y */
        0,0,0,             /* end mark */
    };
    return ilaminar2(dim, x, v);
}


/* Lではない例 (M + M) */
static double f4(int dim, int x[]) {
    double r = f3(dim, x);

    assert(dim == 4);  /* check dimension */
    r +=  2 * x[0] * x[0];
    r +=  3 * x[1] * x[1];
    r +=  2 * (x[2]+5) * (x[2]+5);
    r +=  5 * (x[3]-5) * (x[3]-5);
    return r;
}

static void test(int dim, double f(int,int*), char *f_name,
          int lower[], int upper[]) {
    counter_reset();
    if (lconv_validate(dim, f, lower, upper)) {
        printf("%s(x) is an L-convex func.\n", f_name);
    } else {
        printf("%s(x) is NOT an L-convex func.\n", f_name);
    }
    counter_print();
}

/* L判定の高速化バージョン */
/* 壊れてそうなところの目星をつけるために IFF/FW で最適化してみる */
/* 間違った最適解の周囲をチェックする */
static void test2(int dim, double f(int,int*), char *f_name, int *init) {
    int i;

    int *x     = ivector_new(dim);
    int *lower = ivector_new(dim);
    int *upper = ivector_new(dim);
    ivector_copy(dim, x, init);

    lgconv_minimize_FW(dim, f, x);
    ivector_copy(dim, lower, x);
    ivector_copy(dim, upper, x);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    test(dim, f, f_name, lower, upper);

    lgconv_minimize_IFF(dim, f, x);
    ivector_copy(dim, lower, x);
    ivector_copy(dim, upper, x);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    test(dim, f, f_name, lower, upper);
}

void test_lconv_v(const int dim) {
    int *lower = ivector_new(10 /* dim */,-10, -10, -10, -10);
    int *upper = ivector_new(10 /* dim */, 10,  10,  10,  10);

    test2(4, f4, "f4", lower);

    test(4, f4, "f4", lower, upper);
    return;
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    test(2, f2, "f2", lower, upper);
    lower[0]++;
    lower[1]++;
    test(2, f2, "f2", lower, upper);
    test(2, f1, "f1", lower, upper);
}
